import { createSignal, onCleanup } from "solid-js";
import { applyMovesGenerator, type State } from "../../game.ts";
import { unwrap } from "solid-js/store";
import clone from "../../clone.ts";

export type AnimationState<T> = {
    running: true,
    state: T
} | {
    running: false,
    state: undefined
};

export function useAnimationQueue<T>(
    delay: number = 1000,
    callback: (animationState: AnimationState<T>) => void,
) {
    const generatorQueue: Generator<T, T | undefined, unknown>[] = [];
    const [running, setRunning] = createSignal(false);

    let interval: ReturnType<typeof setInterval> | undefined = undefined;

    function scheduler() {
        if (generatorQueue.length === 0) {
            clearInterval(interval);
            interval = undefined;
            setRunning(false);
            callback({
                running: false,
                state: undefined,
            });
        } else {
            const generator = generatorQueue[0]!;
            const next = generator.next();

            if (next.done) {
                generatorQueue.shift();
            }

            if (next.value !== undefined) {
                callback({
                    running: true,
                    state: next.value
                });
            } else {
                setImmediate(scheduler);
            }
        }
    }

    onCleanup(() => {
        if (interval) {
            clearInterval(interval);
            interval = undefined;
            setRunning(false);
        }
    });

    return {
        start(generator: Generator<T, T | undefined, unknown>) {
            generatorQueue.push(generator);

            if (!interval) {
                interval = setInterval(scheduler, delay);
                setRunning(true);
            }
        },
        running,
    };
}

export default function useAnimation(
    state: State,
    delay: number,
) {
    const [temporaryState, setTemporaryState] = createSignal<State>();
    const {
        start,
        running,
    } = useAnimationQueue<State>(delay, (animationState) => {
        setTemporaryState(clone(animationState.state));
    });

    return {
        start() {
            const stateClone = clone(unwrap(state));
            const generator = applyMovesGenerator(stateClone, true) as Generator<State, undefined, unknown>;
            setTemporaryState(clone(stateClone));
            start(generator);
        },
        running,
        state: temporaryState,
        get(): AnimationState<State> {
            return {
                running: running(),
                state: temporaryState()
            } as AnimationState<State>;
        },
    }
}
