import { createEffect, type Accessor, onCleanup } from "solid-js";
import * as draw from "./draw.ts";
import * as game from "../../game.ts";

export type UseBoardMouse = {
    canvas: Accessor<HTMLCanvasElement | undefined>,
    state: game.State,
    getTransform: Accessor<draw.Transform>,
    setHoveredCell: (cell: [x: number, y: number] | null) => void,
    clickTile: (x: number, y: number) => void,
};

export default function useBoardMouse(options: UseBoardMouse) {
    createEffect(() => {
        const c = options.canvas();

        if (!c) return;

        function getTileBelowMouse(event: MouseEvent): [number, number] | null {
            const transform = options.getTransform();
            const bounds = c!.getBoundingClientRect();

            const x = ((event.clientX - bounds.left) / window.devicePixelRatio - transform.sx) / transform.scale;
            const y = ((event.clientY - bounds.top) / window.devicePixelRatio - transform.sy) / transform.scale;

            const [hexX, hexY] = draw.inverseHexPosition(x, y);
            if (game.cellIn(options.state.cells, hexX, hexY)) {
                return [hexX, hexY];
            } else {
                return null;
            }
        }

        function mouseDown(event: MouseEvent) {
            const tile = getTileBelowMouse(event);
            if (tile) {
                options.clickTile(tile[0], tile[1]);
            }
        }

        function mouseMove(event: MouseEvent) {
            const tile = getTileBelowMouse(event);
            options.setHoveredCell(tile);
        }

        c.addEventListener("mousedown", mouseDown);
        c.addEventListener("mousemove", mouseMove);
        c.addEventListener("mouseleave", () => options.setHoveredCell(null));
        onCleanup(() => {
            c.removeEventListener("mousedown", mouseDown);
            c.removeEventListener("mousemove", mouseMove);
        });
    });
}
