import { createMemo, type Accessor, createResource } from "solid-js";
import * as game from "../../game.js";
import { loadTileImages } from "../tiles.js";
import * as draw from "./draw.js";
import type { Tool } from "../Toolbar.js";
import type { AnimationState } from "./useAnimation.ts";

export type UseDrawBoardOptions = {
    canvas: Accessor<HTMLCanvasElement | undefined>,
    state: game.State,
    getTransform: Accessor<draw.Transform>,
    hoveredCell: Accessor<[x: number, y: number] | null>,
    selectedTool: Accessor<Tool>,
    getAnimationState: Accessor<AnimationState<game.State>>,
    playerID?: string | undefined,
    targetTiles: Accessor<[x: number, y: number][]>,
    gameOver: Accessor<boolean>,
};

export default function useDrawBoard(options: UseDrawBoardOptions) {
    const {
        canvas,
        state,
        getTransform,
        hoveredCell,
        selectedTool,
        getAnimationState,
        playerID,
        targetTiles,
        gameOver
    } = options;

    const [tileImages] = createResource(loadTileImages);

    const cells = createMemo(() => {
        const animationState = getAnimationState();
        if (animationState.running) {
            return animationState.state.cells;
        } else if (playerID) {
            return game.getLocalState(state, playerID);
        } else {
            return state.cells;
        }
    });

    const animationRunning = () => getAnimationState().running;

    return function drawBoard() {
        const ctx = canvas()?.getContext("2d");
        const images = tileImages();
        if (!ctx || !("cells" in state) || !images) return;

        ctx.imageSmoothingQuality = "high";

        ctx.clearRect(0, 0, ctx.canvas.width, ctx.canvas.height);
        const transform = getTransform();
        const transformedCtx = new draw.TransformedCanvas2DCtx(ctx, transform);

        if (hoveredCell() && !animationRunning() && !gameOver()) {
            const [x, y] = hoveredCell()!;
            const tool = selectedTool();
            let valid = true;

            if (playerID) {
                // TODO: pass this logic as a parameter
                if (tool.type === "placeBuilding") {
                    valid = game.canPlaceBuilding(
                        state.cells,
                        game.remainingResources(state.cells, state.resources, state.moves, playerID),
                        playerID,
                        x,
                        y,
                        tool.building
                    );
                } else if (tool.type === "attack" && tool.selected) {
                    valid = !!targetTiles().find(([x2, y2]) => x2 === x && y2 === y);
                } else if (tool.type === "attack" && !tool.selected) {
                    valid = game.canAttackFrom(
                        state.cells,
                        game.remainingResources(state.cells, state.resources, state.moves, playerID),
                        playerID,
                        x,
                        y
                    );
                }
            }
            draw.drawCellOutline(transformedCtx, x, y, valid);
        }

        for (const [x, y, cell] of game.iterateCells(cells())) {
            draw.drawCell(transformedCtx, cell, x, y, images);
            if (cell.active) {
                draw.drawCellOutline(transformedCtx, x, y, true);
            }
            if (cell.target) {
                draw.drawCellTarget(transformedCtx, x, y, images);
            }
        }

        for (const [x, y] of targetTiles()) {
            draw.drawCellTarget(transformedCtx, x, y, images);
        }
    }
}
