import { BASE_URL, PLAYER_COLORS } from "../consts.js";
import classes from "./Toolbar.module.css";
import PlayerTile from "../tile-player-any.svg?raw";
import { Building, Buildings, PlaceableBuildings } from "../game.js";
import { For, Match, Switch } from "solid-js";

const WIDTH = 173.2;
const HEIGHT = 200;

export type Tool = {
    type: "placeBuilding",
    building: keyof typeof Buildings,
} | {
    type: "attack",
    selected: null | {
        x: number,
        y: number,
    }
};

export type ToolbarProps = {
    playerID: string,
    selectedTool: () => Tool,
    setSelectedTool: (tool: Tool) => void,
    resources: () => number,
    resourceGain: () => number,
    setReady: (ready: boolean) => void,
    ready: () => boolean,
};

export default function Toolbar(props: ToolbarProps) {
    function isPlacingBuilding(building: keyof typeof Buildings) {
        const tool = props.selectedTool();
        if (tool.type !== "placeBuilding") return false;
        return tool.building === building;
    }

    function selectBuilding(building: keyof typeof Buildings) {
        props.setSelectedTool({
            type: "placeBuilding",
            building,
        });
    }

    return <nav class={classes.toolbar}>
        <ul>
            <ToolbarItem
                type="attack"
                selected={() => props.selectedTool().type === "attack"}
                onSelect={() => props.setSelectedTool({ type: "attack", selected: null })}
            />
            <For each={Object.values(Buildings).flatMap((definition): PlaceableBuildings[] => "cost" in definition ? [definition.name] : [])}>
                {(name) => {
                    return <ToolbarItem
                        playerID={props.playerID}
                        type={name}
                        selected={() => isPlacingBuilding(name)}
                        onSelect={() => selectBuilding(name)}
                    />
                }}
            </For>
            <li>
                Resources: {props.resources()} (+{props.resourceGain()})
                <br />
                <button onClick={() => props.setReady(!props.ready())}>
                    {props.ready() ? "Waiting..." : "Ready"}
                </button>
            </li>
        </ul>
        <Description selectedTool={props.selectedTool} />
    </nav>;
}


function ToolbarItem(props: {
    playerID?: string,
    type: string,
    selected: () => boolean,
    onSelect: () => void,
}) {
    const color = props.playerID ? PLAYER_COLORS[+props.playerID % PLAYER_COLORS.length] : "#707070";

    return <li
        class={[classes.tool, props.selected() ? classes.selected : ""].filter(Boolean).join(" ")}
        tabindex={0}
        onClick={props.onSelect}
        onKeyDown={(event) => {
            if (event.key === "Enter") props.onSelect();
        }}
        role="button"
    >
        <div
            innerHTML={PlayerTile}
            class={props.type !== "road" ? classes.background : classes["only-background"]}
            style={{
                "--player-color": color
            }}
        ></div>
        {
            props.type !== "road" &&
            <img class={classes.icon} src={`${BASE_URL}/tile-${props.type}.svg`} width={WIDTH} height={HEIGHT} />
        }
    </li>;
}

type KeyOf<T> = T extends object ? keyof T : never;
type TryIndex<T, K extends string> = T extends Record<K, unknown> ? T[K] : never;

type Stats = {
    [K in KeyOf<Building>]:
        (value: TryIndex<Building, K>) => string
}

const STATS = {
    cost: (cost: number) => `Cost: ${cost}`,
    gain: (gain: number) => `Resource gain: ${gain}/turn`,
    hp: (hp: number) => `Health: ${hp}`,
    placedOn: (tiles: readonly (keyof typeof Buildings)[]) => {
        const tileNames = tiles.map(tile => Buildings[tile].humanName);
        if (tileNames.length <= 2) {
            return `Must be placed on ${tileNames.join(" or ")}`;
        } else {
            return `Must be placed on ${tileNames.slice(0, -1).join(", ")} or ${tileNames[tiles.length - 1]}`;
        }
    },
    attack: (attack) => {
        return `Attacks for ${attack.power}, costing ${attack.cost} resources`;
    },
} satisfies Partial<Stats>;

function Description(props: {
    selectedTool: () => Tool
}) {
    const stats = (building: keyof typeof Buildings) => {
        const keys = Object.keys(STATS) as (keyof typeof STATS)[];
        const buildingRules = Buildings[building];

        return keys.flatMap((key) => {
            if (key in buildingRules) {
                // SAFETY: can be trivially proven by applying decidability on `key`.
                return [STATS[key]((buildingRules as Record<string, any>)[key])]
            } else {
                return [];
            }
        })
    };

    return <div class={classes.description}>
        <Switch fallback={<i>No tool selected</i>}>
            <Match when={props.selectedTool().type === "placeBuilding"}>
                {(_) => {
                    const building = () => (props.selectedTool() as (Tool & { type: "placeBuilding" })).building;
                    return (<>
                        <h3>{Buildings[building()].humanName}</h3>
                        <p>{Buildings[building()].description}</p>
                        <ul class={classes.stats}>
                            <For each={stats(building())}>
                                {(item, index) => {
                                    return <li data-index={index()}>{item}</li>;
                                }}
                            </For>
                        </ul>
                    </>);
                }}
            </Match>
            <Match when={props.selectedTool().type === "attack"}>
                {(_) => {
                    return (<>
                        <h3>Attack tool</h3>
                        <p>Select one of your buildings to fire an attack from, and then select where to fire the attack</p>
                    </>);
                }}
            </Match>
        </Switch>
    </div>;
}
