import type { Game } from "boardgame.io";
import { INVALID_MOVE } from "boardgame.io/core";
import hexNeighborhood from "./utils/hexNeighborhood.ts";

// TODO: partial information

export type PlaceableBuildings = {
    [K in keyof typeof Buildings]: typeof Buildings[K] extends { cost: number } ? K : never
}[keyof typeof Buildings];

export type Cell = ({
    owner: null,
} | {
    owner: string,
    building: keyof typeof Buildings,
    hp: number,
    ghost?: boolean,
}) & {
    attacked?: number,
    active?: boolean,
    target?: boolean,
};

export type Move = {
    type: "placeBuilding",
    x: number,
    y: number,
    building: PlaceableBuildings,
} | {
    type: "attack",
    building: keyof typeof Buildings,
    x: number,
    y: number,
    targetX: number,
    targetY: number,
};

export type State = {
    cells: Record<string, Cell>,
    resources: Record<string, number>,
    moves: Record<string, Move[]>
}

export const Buildings = {
    base: {
        name: "base",
        description: "Your HQ, if you lose it, you lose the game!",
        humanName: "",
        gain: 3,
        hp: 5,
    },
    road: {
        name: "road",
        description: "The most basic tile, can be placed on any empty tile adjacent to one of your tiles. Allows buildings to be constructed on it.",
        humanName: "Road",
        cost: 1,
        hp: 1,
    },
    defender: {
        name: "defender",
        description: "Your first line of defense: has good HP and can attack a single adjacent tile to prevent enemy expansion to the border of your territory.",
        humanName: "Defender",
        cost: 2,
        hp: 3,
        placedOn: ["road"],
        attack: {
            power: 1,
            cost: 0,
            maxMoves: 1,
            targets(cells, x, y) {
                const currentPlayer = getCell(cells, x, y)?.owner;
                if (!currentPlayer) return [];
                return [...adjacentCells(x, y)].filter(([x, y]) => {
                    const cell = getCell(cells, x, y);
                    if (cell?.owner) {
                        return cell.owner !== currentPlayer;
                    } else {
                        return true;
                    }
                });
            },
            damageZone(_cells, _x, _y, targetX, targetY) {
                return [[targetX, targetY]];
            }
        }
    },
    laser: {
        name: "laser",
        description: "A long-range defensive unit, able to cause moderate damage to 2 tiles in a row. Does not pierce",
        humanName: "Laser",
        cost: 4,
        hp: 3,
        placedOn: ["road", "defender"],
        attack: {
            power: 2,
            cost: 1,
            maxMoves: 1,
            targets(_cells, x, y) {
                return [...adjacentCells(x, y)];
            },
            damageZone(cells, x, y, targetX, targetY) {
                const dx = targetX - x;
                const dy = targetY - y;
                const MAX_DIST = 2;

                const squares = new Array(MAX_DIST).fill(null).map((_, dist): [number, number] => [x + dx * (dist + 1), y + dy * (dist + 1)]);
                const maxDist = squares.findIndex(([x, y]) => !!getCell(cells, x, y)?.owner);

                return squares.slice(0, maxDist === -1 ? MAX_DIST : maxDist + 1);
            }
        }
    },
    factory: {
        name: "factory",
        description: "Gives a steady income of resources, but will explode the adjacent tiles if destroyed.",
        humanName: "Factory",
        cost: 3,
        gain: 2,
        hp: 2,
        placedOn: ["road"]
    },
    wall: {
        name: "wall",
        description: "A strong defensive unit, that is unable to attack.",
        humanName: "Wall",
        cost: 3,
        hp: 5,
        placedOn: ["road", "defender"]
    },
    artillery: {
        name: "artillery",
        description: "A long-range unit, capable of shooting over defenses for moderate damage up to 3 tiles away.",
        humanName: "Artillery",
        cost: 6,
        hp: 1,
        placedOn: ["road", "defender"],
        attack: {
            cost: 2,
            power: 2,
            maxMoves: 1,
            targets(_cells, x, y) {
                return [
                    ...hexNeighborhood(x, y, 2),
                    ...hexNeighborhood(x, y, 3),
                ];
            },
            damageZone(_cells, _x, _y, targetX, targetY) {
                return [[targetX, targetY]];
            }
        }
    },
} as const satisfies Readonly<Record<string, Readonly<{
    name: string,
    description: string,
    humanName: string,
    cost?: number,
    hp: number,
    gain?: number,
    placedOn?: readonly string[],
    attack?: {
        power: number,
        cost: number,
        maxMoves: number,
        targets: (cells: Record<string, Cell>, x: number, y: number) => [x: number, y: number][],
        damageZone: (cells: Record<string, Cell>, x: number, y: number, targetX: number, targetY: number) => [x: number, y: number][],
    }
}>>>;

export type Building = typeof Buildings[keyof typeof Buildings];

type MapMove<T> = T extends (state: never, ...args: infer Args) => infer Return ? (...args: Args) => Exclude<Return, typeof INVALID_MOVE> : never;
type MapMoves<T extends Record<string, (...args: never[]) => unknown>> = {
    [K in keyof T]: MapMove<T[K]>
}

export type Moves = Partial<MapMoves<{
    attack: typeof attack,
    placeBuilding: typeof placeBuilding,
    setReady: (state: {}) => void,
    setConnected: (state: {}) => void,
}>>;

export type SetupData = {
    size?: number,
    initialResources?: number
};

export const DEFAULT_SETUP_DATA: Required<SetupData> = {
    size: 4,
    initialResources: 2,
}

export const AcrossTheHex: Game<State, Record<string, unknown>, SetupData> = {
    name: "across-the-hex",
    setup({ ctx }, setupData = {}) {
        const { size, initialResources } = {
            ...DEFAULT_SETUP_DATA,
            ...setupData
        };
        const cells = initGrid(size);
        const players = new Array(ctx.numPlayers).fill(null).map((_, id) => id.toString());

        setCell(cells, 0, 0, {
            owner: "0",
            building: "base",
            hp: Buildings.base.hp
        });

        setCell(cells, size * 2 - 2, size * 2 - 2, {
            owner: "1",
            building: "base",
            hp: Buildings.base.hp
        });

        return {
            cells,
            resources: initResources(players, initialResources),
            moves: emptyMoves(players),
        };
    },
    turn: {
        onBegin(ctx) {
            // Dispatch all players into the prepare stage
            ctx.events.setActivePlayers({
                all: "prepare"
            });
        },
        onEnd(ctx) {
            if (!ctx.ctx.activePlayers || !areAllPlayersReady(ctx.ctx.activePlayers)) {
                throw new Error("Assertion error: not all players were ready when turn ended");
            }

            applyMoves(ctx.G);

            if (isGameOver(ctx.G)) {
                ctx.events.endGame();
            }
        },
        endIf(ctx) {
            if (!ctx.ctx.activePlayers) return false;
            return areAllPlayersReady(ctx.ctx.activePlayers);
        },
        stages: {
            prepare: {
                next: "ready",
                moves: {
                    setReady: {
                        move (ctx) {
                            ctx.events.endStage();
                        },
                        ignoreStaleStateID: true,
                    },
                    placeBuilding: {
                        move: placeBuilding,
                        ignoreStaleStateID: true
                    },
                    attack: {
                        move: attack,
                        ignoreStaleStateID: true
                    },
                },
            },
            ready: {
                moves: {}
            }
        }
    }
} as const;

function getCostForBuilding(
    existingCell: Cell,
    building: PlaceableBuildings
): number {
    const buildingRules = Buildings[building];
    if (!existingCell.owner) return buildingRules.cost;

    const existingBuilding = Buildings[existingCell.building];

    if (existingBuilding.name === "road" || !("cost" in existingBuilding)) return buildingRules.cost;

    return buildingRules.cost - existingBuilding.cost;
}

export function canPlaceBuilding(
    cells: Record<string, Cell>,
    resources: number,
    playerID: string,
    x: number,
    y: number,
    building: keyof typeof Buildings
): boolean {
    // Cannot place a building outside the field
    if (!cellIn(cells, x, y)) return false;
    // Must place a building next to an already-established friendly building
    if (!hasAdjacentBuilding(cells, x, y, playerID)) return false;

    // Cannot place a building on an opponent's building
    const existingCell = getCell(cells, x, y)!;
    if (existingCell.owner && existingCell.owner !== playerID) return false;

    const buildingRules = Buildings[building];
    if (!("cost" in buildingRules)) return false;

    if (!("placedOn" in buildingRules)) {
        // Cannot place a building on an existing building without a placedOn property
        if (existingCell.owner) return false;
    } else {
        // Can only place a building on an existing building that is a member of placedOn
        if (!existingCell.owner) return false;
        if (!(buildingRules.placedOn as readonly string[]).includes(existingCell.building)) return false;
    }

    // Cannot place a building without enough resources
    const cost = getCostForBuilding(existingCell, buildingRules.name);
    if (resources < cost) return false;

    return true;
}

function placeBuilding({ G, playerID }: { G: State, playerID: string }, x: number, y: number, building: keyof typeof Buildings) {
    // Prevent placing two buildings at the same place
    if (G.moves[playerID]!.find((move) => {
        return move.type === "placeBuilding" && move.x === x && move.y === y;
    })) return INVALID_MOVE;

    if (!canPlaceBuilding(
        G.cells,
        remainingResources(G.cells, G.resources, G.moves, playerID),
        playerID,
        x,
        y,
        building
    )) return INVALID_MOVE;

    G.moves[playerID]!.push({
        type: "placeBuilding",
        x,
        y,
        // SAFETY: guaranteed by canPlaceBuilding
        building: building as PlaceableBuildings,
    });

    // G.cells[`${y}:${x}`] = {
    //     owner: playerID
    // };
    // G.resources[playerID] -= Buildings[building].cost;

    return;
}

export function canAttackFrom(
    cells: Record<string, Cell>,
    resources: number,
    playerID: string,
    x: number,
    y: number,
): boolean {
    // Can only attack from owned, attack-ready cells
    const building = getBuilding(cells, x, y, playerID);
    if (!building || !("attack" in building)) return false;

    // Cannot attack without the required resources
    if (resources < building.attack.cost) return false;

    return true;
}

// TODO: make attacking moves and building construction mutually exclusive
export function canAttack(
    cells: Record<string, Cell>,
    resources: number,
    playerID: string,
    x: number,
    y: number,
    targetX: number,
    targetY: number,
): boolean {
    // Can only attack from owned, attack-ready cells
    const building = getBuilding(cells, x, y, playerID);
    if (!building || !("attack" in building)) return false;

    // Cannot attack without the required resources
    if (resources < building.attack.cost) return false;

    // Can only attack to one of the listed target cells
    const targets = building.attack.targets(cells, x, y);
    if (!targets.find(([tx, ty]) => tx === targetX && ty === targetY)) return false;

    return true;
}

function attack({ G, playerID }: { G: State, playerID: string }, x: number, y: number, targetX: number, targetY: number) {
    if (!canAttack(
        G.cells,
        remainingResources(G.cells, G.resources, G.moves, playerID),
        playerID,
        x,
        y,
        targetX,
        targetY,
    )) return INVALID_MOVE;

    const building = getBuilding(G.cells, x, y);
    // The type checker isn't happy with us otherwise
    if (!building || !("attack" in building)) return INVALID_MOVE;

    // Cannot attack more than the listed amount of attacks per turn
    const attackingMoves = (G.moves[playerID] ?? []).filter((move) => {
        return move.type === "attack" && move.x === x && move.y === y
    });
    if (attackingMoves.length >= building.attack.maxMoves) return INVALID_MOVE;

    G.moves[playerID]?.push({
        type: "attack",
        x,
        y,
        targetX,
        targetY,
        building: building.name
    });

    return;
}

export function getBuilding(cells: Record<string, Cell>, x: number, y: number, expectedPlayer?: string): Building | null {
    const cell = getCell(cells, x, y);
    if (!cell?.owner) return null;

    if (expectedPlayer && cell.owner !== expectedPlayer) return null;

    return Buildings[cell.building];
}

export function remainingResources(cells: Record<string, Cell>, resources: Record<string, number>, moves: Record<string, Move[]>, playerID: string): number {
    let result = resources[playerID] ?? 0;

    for (const move of moves[playerID] ?? []) {
        if (move.type === "placeBuilding") {
            const existingCell = getCell(cells, move.x, move.y);
            if (existingCell) {
                result -= getCostForBuilding(existingCell, move.building);
            }
        } else if (move.type === "attack") {
            const building = Buildings[move.building];
            if ("attack" in building) {
                result -= building.attack.cost;
            }
        }
    }

    return result;
}

function initGrid(size: number): Record<string, Cell> {
    const res: Record<string, Cell> = {};

    for (let y = 0; y < size; y++) {
        for (let x = 0; x < size + y; x++) {
            res[`${y}:${x}`] = emptyCell();
        }
    }

    for (let y = size; y < size * 2 - 1; y++) {
        for (let x = y - size + 1; x < size * 2 - 1; x++) {
            res[`${y}:${x}`] = emptyCell();
        }
    }

    return res;
}

function emptyCell(): Cell {
    return {
        owner: null
    };
}

function emptyMoves(players: string[]) {
    let res: Record<string, Move[]> = {};

    for (const player of players) {
        res[player] = [];
    }

    return res;
}

function initResources(players: string[], resources: number = 0): Record<string, number> {
    return Object.fromEntries(players.map((id) => [id, resources] as const));
}

export function cellIn(grid: Record<string, Cell>, x: number, y: number): boolean {
    if (x < 0 || y < 0 || !Number.isInteger(x) || !Number.isInteger(y)) return false;
    return `${y}:${x}` in grid;
}

export function getCell(grid: Record<string, Cell>, x: number, y: number): Cell | null {
    if (!cellIn(grid, x, y)) return null;

    return grid[`${y}:${x}`]!;
}

export function setCell(grid: Record<string, Cell>, x: number, y: number, cell: Cell) {
    if (!cellIn(grid, x, y)) return;

    grid[`${y}:${x}`] = cell;
}

function hasAdjacentBuilding(grid: Record<string, Cell>, x: number, y: number, playerID: string): boolean {
    for (const [x2, y2] of adjacentCells(x, y)) {
        const cell = getCell(grid, x2, y2);
        if (cell?.owner === playerID) return true;
    }

    return false;
}

export function areAllPlayersReady(activePlayers: Record<string, string>): boolean {
    for (const player in activePlayers) {
        if (activePlayers[player] !== "ready") return false;
    }
    return true;
}

export function getResourceGain(cells: Record<string, Cell>, player: string) {
    let res = 0;

    for (const [_x, _y, cell] of iterateCells(cells)) {
        if (cell.owner === player) {
            const building = Buildings[cell.building];
            if ("gain" in building) {
                res += building.gain;
            }
        }
    }

    return res;
}

export function* adjacentCells(x: number, y: number): Iterable<[x: number, y: number]> {
    yield [x - 1, y - 1];
    yield [x, y - 1];
    yield [x - 1, y];
    yield [x + 1, y];
    yield [x, y + 1];
    yield [x + 1, y + 1];
}

export function* iterateCells(grid: Record<string, Cell>) {
    const regexp = /^(\d+):(\d+)$/;
    for (let key in grid) {
        const match = regexp.exec(key);
        if (!match) continue;

        yield [+match[2]!, +match[1]!, grid[key]!] as const;
    }
}

function applyMoves(state: State) {
    // Consume `applyMovesGenerator`, which mutates `state` accordingly.
    for (const _ of applyMovesGenerator(state)) {
        // Noop
    }
}

function* animateDamageZone(state: State, attackedSquares: [x: number, y: number][]) {
    for (const [x, y] of attackedSquares) {
        const cell = getCell(state.cells, x, y);
        if (!cell) continue;
        setCell(state.cells, x, y, {
            ...cell,
            target: true
        });
    }

    yield state;

    for (const [x, y] of attackedSquares) {
        const cell = getCell(state.cells, x, y);
        if (!cell) continue;

        const newCell = {...cell};
        delete newCell["target"];

        setCell(state.cells, x, y, newCell);
    }
}

function animateActiveCell<T, U>(state: State, x: number, y: number, generator?: (() => Generator<T, U>) | Generator<T, U>): Generator<T, U>;
function* animateActiveCell<T, U>(state: State, x: number, y: number, generator?: (() => Generator<T, U>) | Generator<T, U>) {
    let cell = getCell(state.cells, x, y);
    if (cell) {
        setCell(
            state.cells,
            x,
            y,
            {
                ...cell,
                active: true,
            }
        );
    }

    if (typeof generator === "function") {
        yield* generator();
    } else if (generator) {
        yield* generator;
    } else {
        yield state;
    }

    cell = getCell(state.cells, x, y);
    if (cell) {
        const newCell = {...cell};
        delete newCell["active"];
        setCell(
            state.cells,
            x,
            y,
            newCell
        );
    }
}

// Progressively applies the moves in `state`, mutating it, and yielding whenever a move was made.
export function* applyMovesGenerator(state: State, animation?: boolean) {
    const players = Object.keys(state.moves);
    const previousCells = {...state.cells};
    const moves = Object.entries(state.moves)
        .flatMap(([player, moves]): [string, Move][] =>
            moves.map((move) => [player, move])
        );
    state.moves = emptyMoves(players);

    const extractedResources = initResources(players, 0);

    // Resource extracting step
    for (const [_x, _y, cell] of iterateCells(state.cells)) {
        if (cell.owner !== null && cell.owner in state.resources) {
            const building = Buildings[cell.building];
            if ("gain" in building) {
                extractedResources[cell.owner] += building.gain;
            }
        }
    }

    // Building placement step
    const placeBuildingMoves: [player: string, move: Move & { type: "placeBuilding" }][] =
        moves.flatMap(([player, move]): [string, Move & { type: "placeBuilding" }][] => move.type === "placeBuilding" ? [[player, move]] : []);

    for (const [x, y, _cell] of iterateCells(state.cells)) {
        const orders = placeBuildingMoves
            .filter(([, move]) => move.x === x && move.y === y)
            .filter(([player, move]) => canPlaceBuilding(
                previousCells,
                state.resources[player] ?? 0,
                player,
                move.x,
                move.y,
                move.building
            ));

        if (orders.length === 0) continue;

        // If two players try to build on the same tile, then nothing happens
        if (orders.length > 1) {
            if (animation) {
                yield* animateActiveCell(state, x, y);
            }
            continue;
        }

        const order = orders[0]!;
        const cost = getCostForBuilding(getCell(previousCells, x, y)!, order[1].building);

        setCell(state.cells, x, y, {
            owner: order[0],
            building: order[1].building,
            hp: Buildings[order[1].building].hp
        });

        state.resources[order[0]] -= cost;

        if (animation) {
            yield* animateActiveCell(state, x, y);
        }
    }

    // Attacking step
    let hasAttacked = false;
    const attackingMoves = moves
        .flatMap(([player, move]): [string, Move & { type: "attack" }][] => move.type === "attack" ? [[player, move]] : []);
    for (const [player, move] of attackingMoves) {
        if (!canAttack(
            previousCells,
            state.resources[player] ?? 0,
            player,
            move.x,
            move.y,
            move.targetX,
            move.targetY
        )) {
            continue;
        }

        const building = Buildings[move.building];
        if (!("attack" in building)) continue;

        const attackedSquares = building.attack.damageZone(previousCells, move.x, move.y, move.targetX, move.targetY);

        for (const [x, y] of attackedSquares) {
            const attackedBuilding = getCell(state.cells, x, y);
            if (!attackedBuilding) continue;

            let attacked = building.attack.power;
            if (attackedBuilding.attacked) {
                attacked += attackedBuilding.attacked;
            }

            setCell(state.cells, x, y, {
                ...attackedBuilding,
                attacked
            });
        }

        if (animation) {
            yield* animateActiveCell(state, move.x, move.y, animateDamageZone(state, attackedSquares));
        }
        hasAttacked = true;
    }

    // Damage step
    for (const [x, y, cell] of iterateCells(state.cells)) {
        const hp = Math.max(cell.owner ? cell.hp : 0, 0) - Math.max(cell.attacked ?? 0, 0);

        if (hp <= 0) {
            setCell(state.cells, x, y, {
                owner: null
            });
        } else if (cell.attacked) {
            const newCell = {...cell} as (Cell & { owner: string });
            delete newCell["attacked"];
            newCell.hp = hp;
            setCell(state.cells, x, y, newCell);
        }
    }

    if (animation && hasAttacked) {
        yield state;
    }

    // Resource gathering step
    for (const player of players) {
        state.resources[player] += extractedResources[player]!;
    }
}

export function getLocalState(state: State, playerID: string): State["cells"] {
    const res: State["cells"] = {
        ...state.cells
    };

    for (const move of state.moves[playerID] ?? []) {
        if (move.type === "placeBuilding") {
            setCell(res, move.x, move.y, {
                owner: playerID,
                building: move.building,
                hp: Buildings[move.building].hp,
                ghost: true,
            });
        }
    }

    // Register attack values
    for (const move of state.moves[playerID] ?? []) {
        if (move.type !== "attack") continue;

        const building = getBuilding(state.cells, move.x, move.y, playerID);
        if (!building || !("attack" in building)) continue;

        const attacks = building.attack.damageZone(state.cells, move.x, move.y, move.targetX, move.targetY);
        for (const [x, y] of attacks) {
            const cell = getCell(res, x, y);
            if (!cell) continue;
            setCell(res, x, y, {
                ...cell,
                attacked: (cell.attacked ?? 0) + building.attack.power
            });
        }
    }

    return res;
}

export function isGameOver(state: State): boolean {
    const playersAlive: string[] = [];
    for (const [_x, _y, cell] of iterateCells(state.cells)) {
        if (cell.owner && cell.building === "base") {
            if (!playersAlive.includes(cell.owner)) playersAlive.push(cell.owner);
        }
    }
    return playersAlive.length <= 1;
}
